/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    ArmarX::Gui
* @author     Mirko Waechter ( mirko.waechter at kit dot edu)
* @date       2012
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#include "ArmarXPlotter.h"
#include "ArmarXPlotterDialog.h"

#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/observers/ObserverObjectFactories.h>
#include <ArmarXCore/observers/variant/DataFieldIdentifier.h>
#include <ArmarXCore/observers/exceptions/local/InvalidChannelException.h>

//ignore errors about extra ; from qwt
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wpedantic"
#include <qwt_plot.h>
#include <qwt_plot_marker.h>
#include <qwt_plot_curve.h>
#include <qwt_plot_panner.h>
#include <qwt_plot_magnifier.h>
#include <qwt_plot_canvas.h>
#include <qwt_legend.h>
#include <qwt_legend_item.h>
#include <qwt_series_data.h>
#include <qwt_thermo.h>
#include <qwt_scale_draw.h>
#pragma GCC diagnostic pop

//QT
#include <QTimer>
#include <QTime>
#include <QSettings>
#include <QDir>
#include <QStackedLayout>
#include <QTextEdit>

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/filesystem/path.hpp>

#include <IceUtil/IceUtil.h>

#include <sstream>


using namespace std;

namespace armarx
{

    ArmarXPlotter::ArmarXPlotter() :
        ArmarXComponentWidgetController(),
        updateInterval(100),
        shownInterval(60),
        startUpTime(QDateTime::currentDateTime())


    {
        setTag("Plotter");
        ui.setupUi(getWidget());
        timer = new QTimer(getWidget());
        //bars = new std::vector<QwtThermo*>();
        barlayout = new QHBoxLayout;
        ////////////////
        //  Setup Plotter
        ///////////////
        // panning with the left mouse button
        (void) new QwtPlotPanner(ui.qwtPlot->canvas());

        // zoom in/out with the wheel
        QwtPlotMagnifier* magnifier = new QwtPlotMagnifier(ui.qwtPlot->canvas());
        magnifier->setAxisEnabled(QwtPlot::xBottom, false);

        ui.qwtPlot->canvas()->setPaintAttribute(QwtPlotCanvas::BackingStore, false); //increases performance for incremental drawing

        QwtLegend* legend = new QwtLegend;
        legend->setItemMode(QwtLegend::CheckableItem);
        ui.qwtPlot->insertLegend(legend, QwtPlot::BottomLegend);


        ui.qwtPlot->setAxisTitle(QwtPlot::xBottom, "Time (in sec)");
        ui.qwtPlot->enableAxis(QwtPlot::yLeft, false);

        ui.qwtPlot->enableAxis(QwtPlot::yRight, true);
        ui.qwtPlot->setAxisAutoScale(QwtPlot::yRight, true);
        //        ui.qwtPlot->setAutoReplot();

        dialog = new ArmarXPlotterDialog(getWidget(), NULL);
        loggingDir = (QDir::currentPath());
        dialog->ui.editLoggingDirectory->setText(loggingDir);

        graphStyle = ui.CBgraphStyle->currentIndex();
        stackedLayout = new QStackedLayout(ui.verticalLayout);
        stackedLayout->addWidget(ui.qwtPlot);

        QWidget* barsWidget = new QWidget(getWidget());
        barsWidget->setLayout(barlayout);
        barsWidget->setAccessibleName("barsWidget");
        stackedLayout->addWidget(barsWidget);

        getWidget()->setLayout(stackedLayout);

    }

    ArmarXPlotter::~ArmarXPlotter()
    {
        //        if(dialog && this->getState() == eManagedIceObjectInitialized)
        //            getArmarXManager()->removeObjectNonBlocking(dialog);
        //        delete dialog;

    }







    void ArmarXPlotter::onInitComponent()
    {
        json = new JSONObject(getIceManager()->getCommunicator());

    }

    void ArmarXPlotter::onConnectComponent()
    {
        ARMARX_VERBOSE << "ArmarXPlotter started" << flush;
        dialog->setIceManager(getIceManager());

        connect(dialog, SIGNAL(accepted()), this, SLOT(configDone()));
        connect(ui.BTNEdit, SIGNAL(clicked()), this, SLOT(configDialog()));
        connect(timer, SIGNAL(timeout()), this, SLOT(updateGraph()));
        connect(ui.BTNPlotterStatus, SIGNAL(toggled(bool)), this, SLOT(plottingPaused(bool)));

        connect(ui.BTNAutoScale, SIGNAL(toggled(bool)), this, SLOT(autoScale(bool)));
        connect(ui.qwtPlot, SIGNAL(legendChecked(QwtPlotItem*, bool)),
                SLOT(showCurve(QwtPlotItem*, bool)));

        connect(ui.btnLogToFile, SIGNAL(toggled(bool)), this, SLOT(toggleLogging(bool)));

        connect(ui.CBgraphStyle, SIGNAL(currentIndexChanged(int)), this, SLOT(onGraphStyleChanged(int)));

        if (!QMetaObject::invokeMethod(this, "setupCurves"))
        {
            ARMARX_WARNING << "Failed to invoke enable";
        }


        usingTopic("TopicReplayerListener");

    }


    void ArmarXPlotter::onExitComponent()
    {
        unsubscribeFromTopic("TopicReplayerListener");

        if (pollingTask)
        {
            pollingTask->stop();
        }

    }

    void ArmarXPlotter::onCloseWidget(QCloseEvent* event)
    {
        ARMARX_VERBOSE << "closing" << flush;
        timer->stop();

        if (logstream.is_open())
        {
            logstream.close();
        }
    }

    void ArmarXPlotter::ButtonAddSensorChannelClicked()
    {
    }

    void ArmarXPlotter::configDialog()
    {
        if (!dialog)
        {
            return;
        }

        dialog->ui.spinBoxUpdateInterval->setValue(updateInterval);
        dialog->ui.spinBoxShownInterval->setValue(shownInterval);
        dialog->ui.syncDataLogging->setChecked(syncDataLogging);
        dialog->ui.listWidget->clear();
        dialog->ui.listWidget->addItems(selectedChannels);
        //        if(dialog->exec())
        //            configDone();
        dialog->setModal(true);
        dialog->show();
        //        dialog->raise();
        //        dialog->activateWindow();
        //        dialog->setParent(0);
    }

    void ArmarXPlotter::configDone()
    {
        {
            ScopedLock lock(dataMutex);
            updateInterval = dialog->ui.spinBoxUpdateInterval->value();
            shownInterval = dialog->ui.spinBoxShownInterval->value();
            pollingInterval = dialog->ui.spinBoxPollingInterval->value();
            selectedChannels = dialog->getSelectedChannels();
            loggingDir = dialog->ui.editLoggingDirectory->text();
            syncDataLogging = dialog->ui.syncDataLogging->isChecked();
            ui.btnLogToFile->setChecked(false);
            ui.BTNPlotterStatus->setChecked(false);
            markers.clear();
            curves.clear();
            dataMap.clear();
            ui.qwtPlot->detachItems();
            clearBarList();
        }

        setupCurves();
    }

    void ArmarXPlotter::toggleLogging(bool toggled)
    {
        std::lock_guard<std::mutex> lock(mutex);

        if (toggled)
        {
            boost::filesystem::path outputPath;
            if (filename.empty())
            {
                outputPath = loggingDir.toStdString() + "/datalog.csv";

                std::string time = IceUtil::Time::now().toDateTime();
                boost::replace_all(time, "/", "-");
                boost::replace_all(time, " ", "_");
                boost::replace_all(time, ":", "-");
                outputPath = outputPath.parent_path() / (outputPath.stem().string() + "-" + time + outputPath.extension().string());

            }
            else
            {
                boost::filesystem::path path = filename;
                outputPath = loggingDir.toStdString() + "/" + path.stem().string() + ".csv";
            }

            ARMARX_INFO << "Logging to " << outputPath.string();
            logstream.open(outputPath.string());
            logStartTime = IceUtil::Time::now();

            if (!logstream.is_open())
            {
                ARMARX_ERROR << "Could not open file for logging: " << outputPath.string();
                ui.btnLogToFile->setChecked(false);
            }
            else
            {
                logstream << "Timestamp";

                for (auto & channel : selectedChannels)
                {
                    logstream <<  "," << channel.toStdString();
                }

                logstream << std::endl;
            }
        }
        else
        {
            filename =  "";
            logstream.flush();
            ARMARX_INFO << "done writing log";
            logstream.close();
        }
    }

    void ArmarXPlotter::onGraphStyleChanged(int idx)
    {
        if (graphStyle != idx)
        {
            graphStyle = idx;
            stackedLayout->setCurrentIndex(graphStyle);

            switch (idx)
            {
                case 0: //curves can have various scalings

                    ui.BTNAutoScale->setDisabled(false);
                    break;

                case 1: //bar chart only autoscale possible

                    ui.BTNAutoScale->setDisabled(true);
                    break;

                default:
                    break;
            }
        }
    }

    void ArmarXPlotter::pollingExec()
    {
        std::map<string, VariantPtr> newData = getData(selectedChannels);


        if (ui.btnLogToFile->isChecked())
        {
            logToFile(IceUtil::Time::now(),  newData);
        }
    }



    void ArmarXPlotter::updateGraph()
    {
        ScopedLock lock(dataMutex);
        IceUtil::Time curTime = TimeUtil::GetTime();
        // erase old markers
        for (auto & markerChannel : markers)
        {
            for (auto it = markerChannel.second.begin(); it != markerChannel.second.end();)
            {
                if ((curTime - it->first).toSecondsDouble() > shownInterval)
                {
                    it = markerChannel.second.erase(it);
                }
                else
                {
                    it++;
                }
            }
        }

        QPointF p;

        //        int size = shownInterval*1000/updateInterval;
        GraphDataMap::iterator it = dataMap.begin();


        for (; it != dataMap.end(); ++it)
        {

            QVector<QPointF> pointList;
            pointList.clear();
            auto datafieldId = it->first;
            std::vector<TimeData>& dataVec = it->second;
            //            int newSize = min(size,(int)dataVec.size());
            pointList.resize(dataVec.size());
            auto addMarker = [this](IceUtil::Time & age, VariantPtr & var, std::string & datafieldId)
            {
                std::map < std::string, boost::shared_ptr<QwtPlotMarker>>& markerSubMap = markers[datafieldId][age];

                std::string value = var->getString();
                auto it2 = markerSubMap.find(value);
                if (it2 != markerSubMap.end())
                {
                    boost::shared_ptr<QwtPlotMarker> marker = it2->second;
                    marker->setXValue(0.001 * age.toMilliSecondsDouble());
                }
                else
                {
                    boost::shared_ptr<QwtPlotMarker> marker(new QwtPlotMarker());
                    marker->setValue(0.001 * age.toMilliSecondsDouble(), 0);
                    marker->setLabel(QwtText(QString::fromStdString(value)));
                    marker->setLineStyle(QwtPlotMarker::VLine);
                    marker->setLabelAlignment(Qt::AlignBottom | Qt::AlignLeft);
                    int i = 0;
                    for (auto & markerSubMap : markers)
                    {
                        if (markerSubMap.first == datafieldId)
                        {
                            break;
                        }
                        i++;
                    }
                    marker->setLinePen(QColor(Qt::GlobalColor((i) % 5 + 13)));
                    marker->attach(ui.qwtPlot);
                    markerSubMap[value] = marker;
                }

            };

            try
            {
                int j = 0;

                for (int i = dataVec.size() - 1; i >= 0 ; --i)
                {
                    TimeData& data = dataVec[i];
                    IceUtil::Time age = (data.time - curTime);

                    if (age.toMilliSecondsDouble() <= shownInterval * 1000)
                    {
                        VariantPtr var = VariantPtr::dynamicCast(data.data);

                        if (var->getInitialized())
                        {
                            float xValue = 0.001 * age.toMilliSecondsDouble();
                            auto type = var->getType();
                            double yValue = 0;
                            if (type == VariantType::Float && var->getFloat() != std::numeric_limits<float>::infinity())
                            {
                                yValue = (var->getFloat());
                            }
                            else if (type == VariantType::Double && var->getDouble() != std::numeric_limits<double>::infinity())
                            {
                                yValue = (var->getDouble());
                            }
                            else if (type == VariantType::Int && var->getInt() != std::numeric_limits<int>::infinity())
                            {
                                yValue = (var->getInt());
                            }
                            else if (type == VariantType::Bool)
                            {
                                yValue = (var->getBool() ? 1 : -1);
                            }
                            else if (type == VariantType::String)
                            {
                                addMarker(age, var, datafieldId);
                            }
                            else
                            {
                                continue;
                            }

                            if ((signed int)j < pointList.size())
                            {
                                pointList[j].setY(yValue);
                                pointList[j].setX(xValue);
                            }
                            else
                            {

                                p.setY(yValue);
                                p.setX(xValue);
                                pointList.push_back(p);
                            }

                            j++;
                        }
                        else
                        {
                            ARMARX_WARNING << deactivateSpam(3) << "uninitialized field: " << it->first;
                        }

                    }
                    else
                    {
                        break;    // data too old from now
                    }
                }
            }
            catch (...)
            {
                handleExceptions();
            }

            QwtSeriesData<QPointF>* pointSeries = new  QwtPointSeriesData(pointList);

            if (pointList.size() > 0 && curves.find(it->first) != curves.end() && bars.find(it->first) != bars.end())
            {
                QwtPlotCurve* curve = curves[it->first];
                curve->setData(pointSeries);
                curve->setYAxis(QwtPlot::yRight);
                QwtThermo* bar = bars[it->first];
                bar->setValue(pointList.first().y());

                bar->setRange(curve->minYValue(), (curve->maxYValue())); //this autoscales the graph
            }


        }


        ui.qwtPlot->setAxisScale(QwtPlot::xBottom, shownInterval * -1, 0.f);

        if (ui.BTNAutoScale->isChecked())
        {
            ui.qwtPlot->setAxisAutoScale(QwtPlot::yRight, true);
        }

        //        ui.qwtPlot->setAxisScale( QwtPlot::yLeft, -1, 1);

        ui.qwtPlot->replot();
    }

    void ArmarXPlotter::showCurve(QwtPlotItem* item, bool on)
    {
        item->setVisible(on);

        QwtLegendItem* legendItem =
            qobject_cast<QwtLegendItem*>(ui.qwtPlot->legend()->find(item));

        if (legendItem)
        {
            legendItem->setChecked(on);
        }

        ui.qwtPlot->replot();
    }

    void ArmarXPlotter::autoScale(bool toggled)
    {
        ARMARX_VERBOSE << "clicked autoscale" << flush;

        ui.qwtPlot->setAxisAutoScale(QwtPlot::yRight, toggled);
        ui.qwtPlot->replot();

    }

    void ArmarXPlotter::plottingPaused(bool toggled)
    {
        __plottingPaused = toggled;

        if (pollingTask)
        {
            pollingTask->stop();
        }

        if (__plottingPaused)
        {
            timer->stop();
        }
        else
        {
            pollingTask = new PeriodicTask<ArmarXPlotter>(this, & ArmarXPlotter::pollingExec, pollingInterval, false, "DataPollingTask", false);
            pollingTask->start();
            timer->start();
        }
    }



    void ArmarXPlotter::saveSettings(QSettings* settings)
    {
        ScopedLock lock(dataMutex);
        settings->setValue("updateInterval", updateInterval);
        settings->setValue("shownInterval", shownInterval);
        settings->setValue("pollingInterval", pollingInterval);
        settings->setValue("selectedChannels", selectedChannels);
        settings->setValue("autoScaleActive", ui.BTNAutoScale->isChecked());
        settings->setValue("syncDataLogging", syncDataLogging);

    }

    void ArmarXPlotter::loadSettings(QSettings* settings)
    {
        ScopedLock lock(dataMutex);
        updateInterval = settings->value("updateInterval", 100).toInt();
        shownInterval = settings->value("shownInterval", 60).toInt();
        pollingInterval = settings->value("pollingInterval", 33).toInt();
        selectedChannels = settings->value("selectedChannels", QStringList()).toStringList();
        ui.BTNAutoScale->setChecked(settings->value("autoScaleActive", true).toBool());
        syncDataLogging = settings->value("syncDataLogging", false).toBool();

        ARMARX_VERBOSE << "Settings loaded" << flush;
    }

    QwtPlotCurve* ArmarXPlotter::createCurve(const QString& label)
    {
        QwtPlotCurve* curve = new QwtPlotCurve(label);
        curve->setRenderHint(QwtPlotItem::RenderAntialiased);
        curve->setPen(QColor(Qt::GlobalColor((curves.size()) % 15 + 7)));
        curve->setStyle(QwtPlotCurve::Lines);
        curve->setPaintAttribute(QwtPlotCurve::CacheSymbols, true);
        curve->setPaintAttribute(QwtPlotCurve::ClipPolygons, true);

        curve->attach(ui.qwtPlot);
        showCurve(curve, true);

        return curve;
    }

    QwtThermo* ArmarXPlotter::createBar(const QString& label)
    {
        //creates a widget containing a bar chart and a label
        //with the fill color of chart same as its curve

        QwtThermo* bar = new QwtThermo(getWidget());
        bar->setAccessibleName(label);
        bar->setFillBrush(QBrush(QColor(Qt::GlobalColor((bars.size() + 7) % 15))));
        bar->setStyleSheet("* { background-color: rgb(240, 240, 240); }"); //neccessary because white is a picked fill color

        QTextEdit* lab = new QTextEdit(label, getWidget());
        lab->setStyleSheet("* { background-color: rgb(240, 240, 240); }");
        lab->setLineWrapMode(QTextEdit::WidgetWidth);
        lab->setTextInteractionFlags(0);
        lab->setFrameStyle(0);
        int linecount = lab->document()->blockCount();
        lab->setMaximumHeight(70 * linecount);
        lab->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
        QVBoxLayout* layout = new QVBoxLayout();
        layout->addWidget(bar);
        layout->addWidget(lab);

        QWidget* widget = new QWidget(layout->widget());
        widget->setLayout(layout);
        barlayout->addWidget(widget);

        return bar;
    }

    void ArmarXPlotter::setupCurves()
    {
        {

            ScopedLock lock(dataMutex);

            markers.clear();
            for (int i = 0; i < selectedChannels.size(); i++)
            {
                try
                {
                    ARMARX_VERBOSE << "Channel: " << selectedChannels.at(i).toStdString() << flush;
                    DataFieldIdentifierPtr identifier = new DataFieldIdentifier(selectedChannels.at(i).toStdString());
                    auto prx = getProxy<ObserverInterfacePrx>(identifier->observerName);

                    VariantPtr var = VariantPtr::dynamicCast(prx->getDataField(identifier));
                    auto type = var->getType();
                    if (type == VariantType::String)
                    {
                        // do nothing for now
                    }
                    else if (VariantType::IsBasicType(type))
                    {
                        QwtPlotCurve* curve = createCurve(selectedChannels.at(i));
                        curves[selectedChannels.at(i).toStdString()] = curve;

                        QwtThermo* bar = createBar(selectedChannels.at(i));
                        bars[selectedChannels.at(i).toStdString()] = bar;
                    }
                    else
                    {
                        auto id = identifier->getIdentifierStr();
                        ARMARX_IMPORTANT << id;
                        auto dict = JSONObject::ConvertToBasicVariantMap(json, var);

                        for (auto e : dict)
                        {
                            if (e.first == "timestamp") // TimedVariants always contain a timestamp field which is irrelevant
                            {
                                continue;
                            }
                            VariantTypeId type = e.second->getType();

                            if (type == VariantType::Double
                                || type == VariantType::Float
                                || type == VariantType::Int)
                            {
                                std::string key = id + "." + e.first;
                                ARMARX_INFO << key << ": " << *VariantPtr::dynamicCast(e.second);
                                QwtPlotCurve* curve = createCurve(QString::fromStdString(key));
                                curves[key] = curve;

                                QwtThermo* bar = createBar(QString::fromStdString(key));
                                bars[key] = bar;
                            }
                        }
                    }
                }
                catch (...)
                {
                    handleExceptions();
                }
            }

            ui.qwtPlot->replot();

        }
        timer->start(updateInterval);

        if (pollingTask)
        {
            pollingTask->stop();
        }

        pollingTask = new PeriodicTask<ArmarXPlotter>(this, & ArmarXPlotter::pollingExec, pollingInterval, false, "DataPollingTask", false);
        pollingTask->start();
    }

    void ArmarXPlotter::clearBarList()
    {
        //maybe comparing new and old selected list better than doing a new one every time?
        QLayoutItem* child;
        int containedItems = barlayout->count() - 1;

        for (int i = containedItems; i >= 0 ; i--) //must be done in this order due to internal implementation of layout item numbering
        {
            child = barlayout->itemAt(i);
            barlayout->removeItem(child);
            delete(child->widget());
        }

        bars.clear();
    }


    std::map<string, VariantPtr> ArmarXPlotter::getData(const QStringList& channels)
    {
        map< std::string, DataFieldIdentifierBaseList > channelsSplittedByObserver;
        foreach(QString channel, channels)
        {
            DataFieldIdentifierPtr identifier = new DataFieldIdentifier(channel.toStdString());
            channelsSplittedByObserver[identifier->getObserverName()].push_back(identifier);
            //            ARMARX_INFO << identifier;
        }
        std::map<std::string, VariantPtr> newData;
        {
            ScopedLock lock(dataMutex);

            // first clear to old entries
            auto now = TimeUtil::GetTime();
            GraphDataMap::iterator itmap = dataMap.begin();

            for (; itmap != dataMap.end(); ++itmap)
            {
                std::vector<TimeData>& dataVec = itmap->second;
                int stepSize = std::max<int>(1, dataVec.size() * 0.01);
                int thresholdIndex = -1;

                for (unsigned int i = 0; i < dataVec.size(); i += stepSize)
                {
                    // only delete if entries are older than 2*showninterval
                    // and delete then all entries that are older than showninterval.
                    // otherwise it would delete items on every call, which would be very slow

                    if ((now - dataVec[i].time).toSecondsDouble() > shownInterval * 2
                        || (thresholdIndex != -1 && (now - dataVec[i].time).toSecondsDouble() > shownInterval)
                       )
                    {
                        thresholdIndex = i;
                    }
                    else
                    {
                        break;
                    }
                }

                if (thresholdIndex != -1)
                {
                    unsigned int offset = std::min((int)dataVec.size(), thresholdIndex);

                    //                ARMARX_IMPORTANT << "Erasing " << offset << " fields";
                    if (offset > dataVec.size())
                    {
                        dataVec.clear();
                    }
                    else
                    {
                        dataVec.erase(dataVec.begin(), dataVec.begin() + offset);
                    }
                }

                //            ARMARX_IMPORTANT << deactivateSpam(5) << "size: " << dataVec.size();
            }

        }

        // now get new data
        IceUtil::Time time = TimeUtil::GetTime();
        map<string, DataFieldIdentifierBaseList >::iterator it = channelsSplittedByObserver.begin();

        try
        {
            for (; it != channelsSplittedByObserver.end(); ++it)
            {
                std::string observerName = it->first;

                if (proxyMap.find(observerName) == proxyMap.end())
                {
                    proxyMap[observerName] = getProxy<ObserverInterfacePrx>(observerName);
                }

                //            QDateTime time(QDateTime::currentDateTime());
                TimedVariantBaseList variants = proxyMap[observerName]->getDataFields(it->second);

                ScopedLock lock(dataMutex);
                //            #    ARMARX_IMPORTANT << "data from observer: " << observerName;
                for (unsigned int i = 0; i < variants.size(); ++i)
                {
                    //                    ARMARX_IMPORTANT << "Variant: " << VariantPtr::dynamicCast(variants[i]);
                    VariantPtr var = VariantPtr::dynamicCast(variants[i]);
                    if (!var->getInitialized())
                    {
                        continue;
                    }
                    auto id = DataFieldIdentifierPtr::dynamicCast(it->second[i])->getIdentifierStr();
                    auto type = var->getType();
                    if (type == VariantType::String)
                    {
                        if (dataMap[id].size() == 0 || dataMap[id].rbegin()->data->getString() != var->getString())
                        {
                            // only insert if changed
                            dataMap[id].push_back(TimeData(time, var));
                            newData[id] = var;
                        }
                    }
                    else if (VariantType::IsBasicType(type))
                    {
                        dataMap[id].push_back(TimeData(time, var));
                        newData[id] = var;
                    }
                    else
                    {
                        auto dict = JSONObject::ConvertToBasicVariantMap(json, var);

                        for (const auto & e : dict)
                        {
                            if (e.first == "timestamp") // TimedVariants always contain a timestamp field which is irrelevant
                            {
                                continue;
                            }
                            std::string key = id + "." + e.first;
                            //                            ARMARX_INFO << key << ": " << *VariantPtr::dynamicCast(e.second);
                            VariantPtr var = VariantPtr::dynamicCast(e.second);
                            auto type = var->getType();
                            if (type == VariantType::String)
                            {
                                // complex contain additional strings often, which cannot be selected right now -> disable strings from complex types
                                //                                if (dataMap[id].size() == 0 || dataMap[id].rbegin()->data->getString() != var->getString())
                                //                                {
                                //                                    // only insert if changed
                                //                                    dataMap[id].push_back(TimeData(time, var));
                                //                                    newData[key] = var;
                                //                                }
                            }
                            else
                            {
                                dataMap[key].push_back(TimeData(time, var));
                                newData[key] = var;
                            }
                        }
                    }

                }

            }
        }
        catch (Ice::NotRegisteredException& e)
        {
            ARMARX_WARNING << deactivateSpam(3) << "Caught Ice::NotRegisteredException: " << e.what();
        }
        catch (exceptions::local::InvalidDataFieldException& e)
        {
            ARMARX_WARNING << deactivateSpam(5) << "Caught InvalidDataFieldException: " << e.what();
        }
        catch (exceptions::local::InvalidChannelException& e)
        {
            ARMARX_WARNING << deactivateSpam(5) << "Caught InvalidChannelException: " << e.what();
        }
        catch (armarx::UserException& e)
        {
            ARMARX_WARNING << deactivateSpam(5) << "Caught UserException: " << e.what() << "\nReason: " << e.reason;
        }
        catch (...)
        {
            ARMARX_WARNING << deactivateSpam(5) << GetHandledExceptionString();
        }

        return newData;
    }

    void ArmarXPlotter::logToFile(const IceUtil::Time& time, const std::map<std::string, VariantPtr>& dataMaptoAppend)
    {

        if (logstream.is_open() && dataMaptoAppend.size() > 0)
        {
            logstream << (time - logStartTime).toMilliSecondsDouble() << ",";

            for (auto & channel : selectedChannels)
            {
                logstream << dataMaptoAppend.at(channel.toStdString())->Variant::getOutputValueOnly();
                if (!selectedChannels.endsWith(channel))
                {
                    logstream << ",";
                }
            }

            logstream  << std::endl;
        }
    }



}
