/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::core::PIDController
 * @author     Mirko Waechter ( mirko dot waechter at kit dot edu )
 * @date       2014
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "PIDController.h"
#include <RobotAPI/libraries/core/math/MathUtils.h>

using namespace armarx;


PIDController::PIDController(float Kp, float Ki, float Kd, double maxControlValue, double maxAcceleration) :
    Kp(Kp),
    Ki(Ki),
    Kd(Kd),
    integral(0),
    derivative(0),
    previousError(0),
    processValue(0),
    target(0),
    controlValue(0),
    maxControlValue(maxControlValue),
    maxAcceleration(maxAcceleration)
{
    reset();
}

void PIDController::reset()
{
    ScopedRecursiveLock lock(mutex);
    firstRun = true;
    previousError = 0;
    integral = 0;
    lastUpdateTime = IceUtil::Time::now();
}

void PIDController::update(double measuredValue, double targetValue)
{
    ScopedRecursiveLock lock(mutex);
    IceUtil::Time now = IceUtil::Time::now();

    if (firstRun)
    {
        lastUpdateTime = IceUtil::Time::now();
    }

    double dt = (now - lastUpdateTime).toSecondsDouble();
    update(dt, measuredValue, targetValue);
    lastUpdateTime = now;
}


void PIDController::update(double measuredValue)
{
    ScopedRecursiveLock lock(mutex);
    update(measuredValue, target);
}

void PIDController::setTarget(double newTarget)
{
    ScopedRecursiveLock lock(mutex);
    target = newTarget;
}

void PIDController::setMaxControlValue(double newMax)
{
    ScopedRecursiveLock lock(mutex);
    maxControlValue = newMax;
}

void PIDController::update(double deltaSec, double measuredValue, double targetValue)
{
    ScopedRecursiveLock lock(mutex);
    processValue = measuredValue;
    target = targetValue;

    double error = target - processValue;

    //double dt = (now - lastUpdateTime).toSecondsDouble();
    //    ARMARX_INFO << deactivateSpam() << VAROUT(dt);
    if (!firstRun)
    {
        integral += error * deltaSec;

        if (deltaSec > 0.0)
        {
            derivative = (error - previousError) / deltaSec;
        }
    }

    firstRun = false;
    double oldControlValue = controlValue;
    controlValue = Kp * error + Ki * integral + Kd * derivative;
    double acc = (controlValue - oldControlValue) / deltaSec;
    if (fabs(acc) > maxAcceleration)
    {
        controlValue = oldControlValue + maxAcceleration * deltaSec * math::MathUtils::Sign(acc);
    }
    controlValue = std::min(fabs(controlValue), maxControlValue) * math::MathUtils::Sign(controlValue);
    ARMARX_DEBUG << deactivateSpam(0.5) << " error: " << error << " cV: " << (controlValue) <<  " i: " << (Ki * integral) << " d: " << (Kd * derivative) << " dt: " << deltaSec;

    previousError = error;
    lastUpdateTime += IceUtil::Time::seconds(deltaSec);

}

double PIDController::getControlValue() const
{
    ScopedRecursiveLock lock(mutex);
    return controlValue;
}


MultiDimPIDController::MultiDimPIDController(float Kp, float Ki, float Kd, double maxControlValue, double maxAcceleration) :
    Kp(Kp),
    Ki(Ki),
    Kd(Kd),
    integral(0),
    derivative(0),
    previousError(0),
    maxControlValue(maxControlValue),
    maxAcceleration(maxAcceleration)
{
    reset();
}

void MultiDimPIDController::update(const double deltaSec, const Eigen::VectorXf& measuredValue, const Eigen::VectorXf& targetValue)
{
    ScopedRecursiveLock lock(mutex);
    processValue = measuredValue;
    target = targetValue;

    double error = (target - processValue).norm();

    //double dt = (now - lastUpdateTime).toSecondsDouble();
    //    ARMARX_INFO << deactivateSpam() << VAROUT(dt);
    if (!firstRun)
    {
        integral += error * deltaSec;

        if (deltaSec > 0.0)
        {
            derivative = (error - previousError) / deltaSec;
        }
    }

    firstRun = false;
    Eigen::VectorXf direction = targetValue; // copy size

    if ((target - processValue).norm() > 0)
    {
        direction = (target - processValue).normalized();
    }
    else
    {
        direction.setZero();
    }
    Eigen::VectorXf oldControlValue;
    if (controlValue.rows() > 0)
    {
        oldControlValue = controlValue;
    }
    else
    {
        oldControlValue = Eigen::VectorXf::Zero(measuredValue.rows());
    }
    controlValue = direction * (Kp * error + Ki * integral + Kd * derivative);

    Eigen::VectorXf accVec = (controlValue - oldControlValue) / deltaSec;
    float maxNewJointAcc = accVec.maxCoeff();
    float minNewJointAcc = accVec.minCoeff();
    maxNewJointAcc = std::max<float>(fabs(minNewJointAcc), fabs(maxNewJointAcc));
    if (maxNewJointAcc > maxAcceleration)
    {
        auto newValue = oldControlValue + accVec * maxAcceleration / maxNewJointAcc * deltaSec;
        ARMARX_DEBUG << deactivateSpam(0.5) << VAROUT(maxAcceleration) << VAROUT(maxNewJointAcc) << VAROUT(controlValue)  << VAROUT(oldControlValue) << VAROUT(newValue);
        controlValue = newValue;
    }


    float max = controlValue.maxCoeff();
    float min = controlValue.minCoeff();
    max = std::max<float>(fabs(min), fabs(max));



    if (max > maxControlValue)
    {
        auto newValue = controlValue  * maxControlValue / max;
        ARMARX_DEBUG << deactivateSpam(0.5) << " Control value to big: " << controlValue << " max value: " << maxControlValue << " new value: " << newValue;
        controlValue = newValue;
    }
    ARMARX_DEBUG << deactivateSpam(0.5) << " error: " << error << " cV: " << (controlValue) <<  " i: " << (Ki * integral) << " d: " << (Kd * derivative) << " dt: " << deltaSec;

    previousError = error;
    lastUpdateTime += IceUtil::Time::seconds(deltaSec);

}

void MultiDimPIDController::update(const Eigen::VectorXf& measuredValue, const Eigen::VectorXf& targetValue)
{
    ScopedRecursiveLock lock(mutex);
    IceUtil::Time now = IceUtil::Time::now();

    if (firstRun)
    {
        lastUpdateTime = IceUtil::Time::now();
    }

    double dt = (now - lastUpdateTime).toSecondsDouble();
    update(dt, measuredValue, targetValue);
    lastUpdateTime = now;
}

const Eigen::VectorXf& MultiDimPIDController::getControlValue() const
{
    return controlValue;
}

void MultiDimPIDController::reset()
{
    ScopedRecursiveLock lock(mutex);
    firstRun = true;
    previousError = 0;
    integral = 0;
    lastUpdateTime = IceUtil::Time::now();
    //    controlValue.setZero();
    //    processValue.setZero();
    //    target.setZero();
}

void MultiDimPIDController::setMaxControlValue(double value)
{
    ScopedRecursiveLock lock(mutex);
    maxControlValue = value;
}
